// sndvolc Copyright (c)2002 Tomoya Tokairin
// license: GPL

#include <windows.h>
#include <stdio.h>

#define VER "sndvolc v0.5 7/22/2002 (c)2002 Tomoya Tokairin"

#define MAX_LINECONTROLS 16
#define MAX_CHN 10

//---------------------------------------------------------
// get MIXERLINE of the recording controler
BOOL GetRecControler(HMIXER hMixer,MIXERCAPS mixerCaps,MIXERLINE* lpMixerLine){

	MIXERLINE mixerLine;
	DWORD i;

	for (i = 0;i < mixerCaps.cDestinations; i++)
	{
		memset(&mixerLine,0,sizeof(MIXERLINE));
		mixerLine.cbStruct = sizeof(MIXERLINE);
		mixerLine.dwDestination = i;

		if(mixerGetLineInfo(
			(HMIXEROBJ)hMixer,
			&mixerLine, 
			MIXER_OBJECTF_HMIXER | MIXER_GETLINEINFOF_DESTINATION 
			) != MMSYSERR_NOERROR) return FALSE;
		
		if(mixerLine.dwComponentType == MIXERLINE_COMPONENTTYPE_DST_WAVEIN)
		{
			*lpMixerLine = mixerLine;
			return TRUE;
		}
	}

	return FALSE;
}



//---------------------------------------------------------
// get MIXERCONTROL of MUX
BOOL GetMUX(HMIXER hMixer,
				   MIXERLINE mixerLine, // of recording controler
				   MIXERCONTROL* lpMixerCtrl
				   )
{
	MIXERLINECONTROLS mixerLineCtrl;
	MIXERCONTROL mixerCtrl[MAX_LINECONTROLS];

	DWORD i,dwType;

	memset(&mixerLineCtrl,0,sizeof(MIXERLINECONTROLS));
	mixerLineCtrl.cbStruct = sizeof(MIXERLINECONTROLS);
	mixerLineCtrl.dwLineID = mixerLine.dwLineID;
	mixerLineCtrl.cControls = mixerLine.cControls;
	mixerLineCtrl.cbmxctrl = sizeof(MIXERCONTROL);
	mixerLineCtrl.pamxctrl = mixerCtrl;
	
	if(mixerGetLineControls(
		(HMIXEROBJ)hMixer,
		&mixerLineCtrl, 
		MIXER_OBJECTF_HMIXER 
		| MIXER_GETLINECONTROLSF_ALL // get all controls
		)
		!= MMSYSERR_NOERROR) return FALSE;
	
	for(i = 0; i < mixerLineCtrl.cControls; i++ )
	{
		dwType = mixerCtrl[i].dwControlType;
		if(dwType == MIXERCONTROL_CONTROLTYPE_MUX 
			|| dwType == MIXERCONTROL_CONTROLTYPE_MIXER
			){
			*lpMixerCtrl = mixerCtrl[i];
			return TRUE;
		}
	}

	return FALSE;
}


//---------------------------------------------------------
// get list of controls
BOOL GetListOfControls(HMIXER hMixer,
	  MIXERCONTROL mixerCtrl, // of MUX
	  MIXERCONTROLDETAILS_LISTTEXT* lpMixerCtrlList
	  ){
	
	MIXERCONTROLDETAILS	mixerCtrlDetails;
	MIXERCONTROLDETAILS_LISTTEXT mixerCtrlList[MAX_LINECONTROLS];

	DWORD i;
	
	memset(&mixerCtrlDetails,0,sizeof(MIXERCONTROLDETAILS));
	mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
	mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
	mixerCtrlDetails.cChannels = 1;
	mixerCtrlDetails.cMultipleItems	= mixerCtrl.cMultipleItems;
	mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_LISTTEXT);
	mixerCtrlDetails.paDetails = &mixerCtrlList;
	
	// get the list of controls
	if(mixerGetControlDetails((HMIXEROBJ)hMixer,
		&mixerCtrlDetails,
		MIXER_OBJECTF_HMIXER 
		| MIXER_GETCONTROLDETAILSF_LISTTEXT 
		) != MMSYSERR_NOERROR) return FALSE;
	
	for(i=0;i<mixerCtrl.cMultipleItems;i++) lpMixerCtrlList[i] = mixerCtrlList[i];
	
	return TRUE;
}



//---------------------------------------------------------
// get MIXERCONTROLDETAILS of MUX
BOOL GetStausOfMUX(HMIXER hMixer,
	  MIXERCONTROL mixerCtrl, // of MUX
	  MIXERCONTROLDETAILS* lpMixerCtrlDetails,
	  MIXERCONTROLDETAILS_BOOLEAN* mixerCtrlDetBool
	  ){
	
	MIXERCONTROLDETAILS	mixerCtrlDetails;

	memset(&mixerCtrlDetails,0,sizeof(MIXERCONTROLDETAILS));
	mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
	mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
	mixerCtrlDetails.cChannels = 1;
	mixerCtrlDetails.cMultipleItems	= mixerCtrl.cMultipleItems;
	mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_BOOLEAN);
	mixerCtrlDetails.paDetails = mixerCtrlDetBool;
	
	if(mixerGetControlDetails((HMIXEROBJ)hMixer,
		&mixerCtrlDetails,
		MIXER_OBJECTF_HMIXER 
		| MIXER_GETCONTROLDETAILSF_VALUE
		) != MMSYSERR_NOERROR) return FALSE;
	
	*lpMixerCtrlDetails = mixerCtrlDetails;

	return TRUE;
}


//----------------------------------------------------------------------
// get MIXERCONTROLDETAILS and volume of each control
BOOL GetVolume(HMIXER hMixer,
			   MIXERLINE mixerLine_rec, // of recording controler
			   MIXERCONTROLDETAILS_LISTTEXT* lpMixerCtrlList, // of MUX
			   MIXERCONTROLDETAILS* lpMixerCtrlDetails,
			   DWORD dwVolume[MAX_LINECONTROLS][MAX_CHN],
			   LPDWORD lpdwBound)
{
	
	MIXERLINE mixerLine;
	MIXERCONTROL mixerCtrl;
	MIXERLINECONTROLS mixerLineCtrl;

	MIXERCONTROLDETAILS	mixerCtrlDetails;

	DWORD i;
	
	for(i = 0; i < mixerLine_rec.cConnections; i++)
	{
		mixerLine = mixerLine_rec;
		mixerLine.dwLineID = lpMixerCtrlList[i].dwParam1;
		
		if(mixerGetLineInfo(
			(HMIXEROBJ)hMixer,
			&mixerLine,
			MIXER_OBJECTF_HMIXER | MIXER_GETLINEINFOF_LINEID) != MMSYSERR_NOERROR) return FALSE;
		
		memset(&mixerLineCtrl,0,sizeof(MIXERLINECONTROLS));
		mixerLineCtrl.cbStruct = sizeof(MIXERLINECONTROLS);
		mixerLineCtrl.dwLineID = mixerLine.dwLineID;
		mixerLineCtrl.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
		mixerLineCtrl.cControls = 1;
		mixerLineCtrl.cbmxctrl = sizeof(MIXERCONTROL);
		mixerLineCtrl.pamxctrl = &mixerCtrl;
		
		if(mixerGetLineControls(
			(HMIXEROBJ)hMixer,
			&mixerLineCtrl, 
			MIXER_OBJECTF_HMIXER | MIXER_GETLINECONTROLSF_ONEBYTYPE 
			)
			!= MMSYSERR_NOERROR) return FALSE;

		lpdwBound[0] = mixerCtrl.Bounds.dwMinimum;
		lpdwBound[1] = mixerCtrl.Bounds.dwMaximum;
		
		memset(&mixerCtrlDetails,0,sizeof(MIXERCONTROLDETAILS));
		mixerCtrlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
		mixerCtrlDetails.dwControlID = mixerCtrl.dwControlID;
		mixerCtrlDetails.cChannels = mixerLine.cChannels;
		mixerCtrlDetails.cMultipleItems	= mixerCtrl.cMultipleItems;
		mixerCtrlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
		mixerCtrlDetails.paDetails = dwVolume[i];

		if(mixerGetControlDetails((HMIXEROBJ)hMixer,
			&mixerCtrlDetails,
			MIXER_OBJECTF_HMIXER 
			| MIXER_GETCONTROLDETAILSF_VALUE
			) != MMSYSERR_NOERROR) return FALSE;
		
		lpMixerCtrlDetails[i] = mixerCtrlDetails;
	}
	
	return TRUE;
}



//---------------------------------------------------------
// main
int main(int argc, char* argv[]){

	DWORD dwMixerID,dwMixNum,dwControl;
	HMIXER hMixer = NULL;
	MIXERCAPS mixerCaps;

	MIXERLINE mixerLine_rec; // for recording controler

	// for MUX (or MIXER)
	MIXERCONTROL mixerCtrl_mux; 
	MIXERCONTROLDETAILS	mixerCtrlDetails_mux; 
	MIXERCONTROLDETAILS_BOOLEAN mixerCtrlDetBool[MAX_LINECONTROLS]; // on or off
	MIXERCONTROLDETAILS_LISTTEXT mixerCtrlList[MAX_LINECONTROLS]; // names of controls
	
	// for volume control
	MIXERCONTROLDETAILS mixerCtrlDetails_vol[MAX_LINECONTROLS];
	DWORD dwVolume[MAX_LINECONTROLS][MAX_CHN];
	DWORD dwBound[2];

	DWORD i,i2,dwFoo;
	char szErr[256];

	szErr[0] = '\0';
	fprintf(stderr,"%s\n\n",VER);
	if(argc == 1){
		fprintf(stdout,"sndvolc id [control] [vol1] [vol2] ...\n\n");
	}

	dwMixNum = mixerGetNumDevs();
    if(dwMixNum == 0) goto L_ERR;

	fprintf(stderr,"mixers:\n");
	for(i=0;i < dwMixNum; i++){
		if(mixerGetDevCaps(i,&mixerCaps,sizeof(MIXERCAPS)) != MMSYSERR_NOERROR){
			strcpy(szErr,"mixerGetDevCaps");
			goto L_ERR;
		}
		fprintf(stderr,"[%2d] %s\n",i,mixerCaps.szPname);
	}

	if(argc == 1) goto L_EXIT;

	// get caps
	dwMixerID = atoi(argv[1]);
	if(dwMixerID >= dwMixNum) dwMixerID = 0;
	if(mixerGetDevCaps(dwMixerID,&mixerCaps,sizeof(MIXERCAPS)) != MMSYSERR_NOERROR){
		strcpy(szErr,"mixerGetDevCaps");
		goto L_ERR;
	}

	// open mixer
    if(mixerOpen(&hMixer,dwMixerID,0,0,MIXER_OBJECTF_MIXER) != MMSYSERR_NOERROR){
		strcpy(szErr,"mixerOpen");
		goto L_ERR;
	}
	fprintf(stderr,"\n%s\n",mixerCaps.szPname);

	// get recording control
	if(!GetRecControler(hMixer,mixerCaps,&mixerLine_rec)){
		strcpy(szErr,"GetRecControler");
		goto L_ERR;
	}
	fprintf(stderr,"%s\n",mixerLine_rec.szName);

	// get MUX
	if(!GetMUX(hMixer,mixerLine_rec,&mixerCtrl_mux)){
		strcpy(szErr,"GetMUX");
		goto L_ERR;
	}
	fprintf(stderr,"%s\n",mixerCtrl_mux.szName);

	// get list of controls
	if(!GetListOfControls(hMixer,mixerCtrl_mux,mixerCtrlList)){
		strcpy(szErr,"GetListOfControls");
		goto L_ERR;
	}
	
	if(argc > 2){
		dwControl = min((DWORD)atoi(argv[2]),mixerCtrl_mux.cMultipleItems-1);

		// choose control
		if(!GetStausOfMUX(hMixer,mixerCtrl_mux,&mixerCtrlDetails_mux,mixerCtrlDetBool)) goto L_ERR;
		memset(mixerCtrlDetBool,0,sizeof(MIXERCONTROLDETAILS_BOOLEAN)*MAX_LINECONTROLS);
		mixerCtrlDetBool[dwControl].fValue = 1;
		if(mixerSetControlDetails((HMIXEROBJ)hMixer,
			&mixerCtrlDetails_mux,
			MIXER_OBJECTF_HMIXER 
			| MIXER_SETCONTROLDETAILSF_VALUE 
			) != MMSYSERR_NOERROR) return FALSE;

		// set volume
		if(argc >= 4){
			if(!GetVolume(hMixer,mixerLine_rec,mixerCtrlList,mixerCtrlDetails_vol,dwVolume,dwBound)) goto L_ERR;
			for(i2=0;i2<mixerCtrlDetails_vol[dwControl].cChannels;i2++){
				if(3+i2 < (DWORD)argc) dwFoo = (DWORD)atoi(argv[3+i2]);
				fprintf(stderr,"[%d %d]\n",mixerCtrlDetails_vol[dwControl].cChannels,dwFoo);
				dwVolume[dwControl][i2] = min(dwFoo,dwBound[1]);
			}
			if(mixerSetControlDetails((HMIXEROBJ)hMixer,
			&mixerCtrlDetails_vol[dwControl],
			MIXER_OBJECTF_HMIXER 
			| MIXER_SETCONTROLDETAILSF_VALUE 
			) != MMSYSERR_NOERROR) return FALSE;
		}
	}

	// show status
	if(!GetStausOfMUX(hMixer,mixerCtrl_mux,&mixerCtrlDetails_mux,mixerCtrlDetBool)) goto L_ERR;
	if(!GetVolume(hMixer,mixerLine_rec,mixerCtrlList,mixerCtrlDetails_vol,dwVolume,dwBound)) goto L_ERR;
	for(i=0;i<mixerCtrl_mux.cMultipleItems;i++){
		fprintf(stderr,"[%2d] ",i);
		if(mixerCtrlDetBool[i].fValue) fprintf(stderr,"o ");
		else  fprintf(stderr,"  ");

		for(i2=0;i2<mixerCtrlDetails_vol[i].cChannels;i2++){
			fprintf(stderr,"%5d ",dwVolume[i][i2]);
		}

		fprintf(stderr,"%s\n",mixerCtrlList[i].szName);
	}

	fprintf(stderr,"\n");

L_EXIT:

    if(hMixer) mixerClose(hMixer);

	return 0;

L_ERR:

    if(hMixer) mixerClose(hMixer);
	fprintf(stderr,"error in %s\n",szErr);

	return 1;
}

// EOF
